#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <fcntl.h>
#include <vector>

using namespace std;

static const unsigned XDIM = 512;  //800;       //700
static const unsigned YDIM = XDIM*9/16; //512

#define INITIAL_W 0
#define USEMANDMAP 0
#define DRAWDOUBLEPIX 0
#define DISPLAYRENDER 1
#define PROFILEMINMAX 0
#define SPHERETEST 0
#define MAXITER 5
#define SKYKLUDGE 0
#define BEAMSPEED 0.01
#define PRECISION 0.00001
#define DISTORTION DEG2RAD(0.05)
#define PROJSIZEX (XDIM/(double)YDIM)
#define PROJSIZEY 1.0
#define PROJDIST 2.4
#define EYEDIST  5.0
#define MAXTRACE     ((EYEDIST+2)/BEAMSPEED)
#define MAXTRACESTR "((EYEDIST+2)/BEAMSPEED)"
#define MIRRORCOUNT 3
#define DIRECTLIGHTMULTIPLIER 0.7
#define INDIRECTLIGHTMULTIPLIER 0.0
#define MIRRORLIGHTMULTIPLIER  0.9
#define PIGMENT     (0.9, 0.2, 0.7)
#define PIGMENTSTR "(0.9, 0.2, 0.7)"
#define AMBIENT     (0.0, 0.0, 0.0)
#define AMBIENTSTR "(0.0, 0.0, 0.0)"
#define TRACEPOW 19
#define SKYPOW 14

#define DEG2RAD(r) ((r) * M_PI / 180.0)

#define FLOAT_ARG const double &

static int ALKUY=0;
#include <linux/fb.h>
static class Video
{
#if DISPLAYRENDER
    int fd;
    unsigned char *p;
    unsigned xres;
public:
    Video() : xres(1504)
    {
        fd = open("/dev/fb0", O_RDWR);
        
        struct fb_cmap cmap;
        cmap.start=0;
        cmap.len  =256;
        cmap.red  =(__u16 *)calloc(256, 2);
        cmap.green=(__u16 *)calloc(256, 2);
        cmap.blue =(__u16 *)calloc(256, 2);
        cmap.transp=NULL;

        ioctl(fd, FBIOGETCMAP, &cmap);
        /* This palette is RGB and used by XAWTV ;) */
        const int r=5;
        const int g=9;
        const int b=5;
        const int rs = 256 / (r - 1);
        const int gs = 256 / (g - 1);
        const int bs = 256 / (b - 1);
        for (int i = 0; i < r*g*b; i++)
        {
            cmap.green[i+16] = (gs * ((i / (r * b)) % g)) * 255;
            cmap.red[i+16]   = (rs * ((i / b) % r)) * 255;
            cmap.blue[i+16]  = (bs * ((i) % b)) * 255;
        }
        ioctl(fd, FBIOPUTCMAP, &cmap);

        free(cmap.red);
        free(cmap.green);
        free(cmap.blue);

        p = (unsigned char *)mmap(NULL, xres*1024, PROT_WRITE, MAP_SHARED, fd, 0);
        if((int)p==-1)
        {
            perror("mmap");
            throw;
        }
    }
    ~Video()
    {
        munmap(p, xres*1024);
    }
    void operator() (unsigned x, unsigned y, unsigned char c)
    {
        y += ALKUY;
        #if DRAWDOUBLEPIX
        if(XDIM<1280/2)x+=(1280-XDIM*2)/4;
        p[(y*2  )*xres+(x*2  )] = c;
        p[(y*2  )*xres+(x*2+1)] = c;
        p[(y*2+1)*xres+(x*2  )] = c;
        p[(y*2+1)*xres+(x*2+1)] = c;
        #else
        p[y*xres+x] = c;
        #endif
    }
#endif
} PSet;

class XYZ
{
public:
    double x,y,z;
#if SKYKLUDGE
    bool nowhere;
#endif
    XYZ() : x(0),y(0),z(0)
#if SKYKLUDGE
    , nowhere(true)
#endif
    {
    }
    XYZ(const XYZ &p) : x(p.x), y(p.y), z(p.z)
#if SKYKLUDGE
    , nowhere(p.nowhere)
#endif
    {
    }
    
    XYZ(FLOAT_ARG xx,FLOAT_ARG yy,FLOAT_ARG zz) : x(xx),y(yy),z(zz)
#if SKYKLUDGE
    , nowhere(false)
#endif
    {
    }
    
    const XYZ &operator +=(const XYZ &p) { x+=p.x; y+=p.y; z+=p.z; return *this; }
    const XYZ &operator -=(const XYZ &p) { x-=p.x; y-=p.y; z-=p.z; return *this; }
    const XYZ &operator /=(FLOAT_ARG p) { x/=p; y/=p; z/=p; return *this; }
    const XYZ &operator *=(FLOAT_ARG p) { x*=p; y*=p; z*=p; return *this; }
    const XYZ operator +(const XYZ &p) const { return XYZ(x+p.x, y+p.y, z+p.z); }
    const XYZ operator -(const XYZ &p) const { return XYZ(x-p.x, y-p.y, z-p.z); }
    const XYZ operator -() const { return XYZ(-x, -y, -z); }
    const XYZ operator /(FLOAT_ARG p) const { return XYZ(x/p, y/p, z/p); }
    const XYZ operator *(FLOAT_ARG p) const { return XYZ(x*p, y*p, z*p); }
    const XYZ &operator= (const XYZ &p)
    {
        x = p.x;
        y = p.y;
        z = p.z;
#if SKYKLUDGE
        nowhere = p.z;
#endif
        return *this;
    }
    
    const double dist(const XYZ &p)const 
    {
        XYZ d(x-p.x, y-p.y, z-p.z);
        return sqrt(d.x*d.x + d.y*d.y + d.z*d.z);
    }
    const double length() const
    {
        return sqrt(x*x + y*y + z*z);
    }
    
    void Normalize()
    {
    	*this /= length();
    }
    void SetLength(FLOAT_ARG newlen)
    {
    	*this *= (newlen / length());
    }
    
    template<class T>
    inline void Rotate(T &m)
    {
        m.Rotate(*this);
    }
    template<class T>
    const XYZ RotatedBy(T &m) const
    {
        XYZ tmp = *this;
        m.Rotate(tmp);
        return tmp;
    }
    const XYZ RotatedAround(const XYZ &axis, FLOAT_ARG angle) const;
    const XYZ MirroredAround(const XYZ &axis) const
    {
        // This could be replaced with RotatedAround(axis, DEG2RAD(180)),
        // but as sin(180)=0 and cos(180)=-1, this is simplified a lot.
        // Terribly lot. :)
        XYZ N(axis);
        N.Normalize();
        return N * (DotProduct(N) * 2.0) - *this;
    }
    const XYZ CrossProduct(const XYZ &p)const
    {
        return XYZ(y*p.z - z*p.y,
                   z*p.x - x*p.z,
                   x*p.y - y*p.x);
    }
    const double DotProduct(const XYZ &p)const
    {
    	return x*p.x + y*p.y + z*p.z;
    }
    
    void print() const
    {
        printf("<%.8f, %.8f, %.8f> (length %.8f)", x,y,z, length());
    }
};



class Matrix
{
    double val[4][4];
public:
    Matrix()
    {
        Zero();
    }
    Matrix(FLOAT_ARG ax, FLOAT_ARG ay, FLOAT_ARG az)
    {
        InitRotate(ax, ay, az);
    }
    Matrix(XYZ axis, FLOAT_ARG ang)
    {
        InitAxisRotate(axis, ang);
    }
    void Zero()
    {
        memset(&val, 0, sizeof val);
    }
    void Identity()
    {
        for(unsigned i=0; i<4; ++i)
            for(unsigned j=0; j<4; ++j)
                val[i][j] = (j==i) ? 1.0 : 0.0;
    }
    void InitRotate(FLOAT_ARG ax, FLOAT_ARG ay, FLOAT_ARG az)
    {
        double tmp[3][4][4];
        
        Identity();

        memset(&tmp, 0, sizeof(tmp));

        tmp[0][1][1]=tmp[0][2][2]=cos(ax); tmp[0][2][1]=-(tmp[0][1][2]=sin(ax));
        tmp[1][0][0]=tmp[1][2][2]=cos(ay); tmp[1][0][2]=-(tmp[1][2][0]=sin(ay));
        tmp[2][0][0]=tmp[2][1][1]=cos(az); tmp[2][1][0]=-(tmp[2][0][1]=sin(az));

        tmp[0][0][0] = 1;
        tmp[0][3][3] = 1;
        tmp[1][1][1] = 1;
        tmp[1][3][3] = 1;
        tmp[2][2][2] = 1;
        tmp[2][3][3] = 1;

        for(int i=0; i<4; i++)
        {
            double m2[4];

            for(int j=0; j<4; j++)
                m2[j] = tmp[1][i][0] * tmp[0][0][j] +
                        tmp[1][i][1] * tmp[0][1][j] +
                        tmp[1][i][2] * tmp[0][2][j] +
                        tmp[1][i][3] * tmp[0][3][j];

            for(int j=0; j<4; j++)
                val[i][j] = m2[0] * tmp[2][0][j] +
                            m2[1] * tmp[2][1][j] +
                            m2[2] * tmp[2][2][j] +
                            m2[3] * tmp[2][3][j];
        }
    }
    /* Math borrowed from Povray, matrices.c:Compute_Axis_Rotation_Transform */
    void InitAxisRotate(XYZ V, FLOAT_ARG ang)
    {
        V.Normalize();
        Identity();

        const double cosx = cos(ang);
        const double sinx = sin(ang);
        
        val[0][0] = V.x * V.x + cosx * (1.0 - V.x * V.x);
        val[0][1] = V.x * V.y * (1.0 - cosx) + V.z * sinx;
        val[0][2] = V.x * V.z * (1.0 - cosx) - V.y * sinx;
        
        val[1][0] = V.x * V.y * (1.0 - cosx) - V.z * sinx;
        val[1][1] = V.y * V.y + cosx * (1.0 - V.y * V.y);
        val[1][2] = V.y * V.z * (1.0 - cosx) + V.x * sinx;
        
        val[2][0] = V.x * V.z * (1.0 - cosx) + V.y * sinx;
        val[2][1] = V.y * V.z * (1.0 - cosx) - V.x * sinx;
        val[2][2] = V.z * V.z + cosx * (1.0 - V.z * V.z);
    }
    
    void Rotate(XYZ &p)
    {
        register double tx = p.x*val[0][0] + p.y*val[1][0] + p.z*val[2][0] + val[3][0];
        register double ty = p.x*val[0][1] + p.y*val[1][1] + p.z*val[2][1] + val[3][1];
        p.z                = p.x*val[0][2] + p.y*val[1][2] + p.z*val[2][2] + val[3][2];
        p.y = ty;
        p.x = tx;
    }
    const Matrix operator * (const Matrix &m) const
    {
        Matrix res;
        for(unsigned i=0; i<3; ++i)
            for(unsigned j=0; j<4; ++j)
                for(unsigned k=0; k<4; ++k)
                    res.val[i][j] += val[i][k] * m.val[k][j];
        res.val[3][3] = 1;
        return res;
    }
    const Matrix &operator *= (const Matrix &m)
    {
        return (*this = *this * m);
    }
};

const XYZ XYZ::RotatedAround(const XYZ &axis, FLOAT_ARG angle) const
{
	Matrix m(axis, angle);
	return RotatedBy(m);
}

class RGB
{
    double r,g,b;
    
    inline const int intensity(double elem) const
    {
        if(elem < 0)return 0;
        const double maxi = max(max(r,g), b);
        if(maxi <= 1)return int(255*elem);
        if(elem > 1)return 255;
        const double jaettavaa = maxi-1;
        int v = int((elem + jaettavaa)*255);
        return v>=255 ? 255 : v;
    }
    
public:
    RGB() : r(0), g(0), b(0)
    {
    }
    RGB(double R, double G, double B) : r(R), g(G), b(B)
    {
    }
    const RGB &operator *=(const RGB &p) { r*=p.r; g*=p.g; b*=p.b; return *this; }
    const RGB &operator +=(const RGB &p) { r+=p.r; g+=p.g; b+=p.b; return *this; }
    const RGB &operator -=(const RGB &p) { r-=p.r; g-=p.g; b-=p.b; return *this; }
    const RGB &operator *=(FLOAT_ARG p) { r*=p; g*=p; b*=b; return *this; }
    const RGB &operator /=(FLOAT_ARG p) { r/=p; g/=p; b/=b; return *this; }
    const RGB operator *(const RGB &p) { return RGB(r*p.r, g*p.g, b*p.b); }
    const RGB operator +(const RGB &p) { return RGB(r+p.r, g+p.g, b+p.b); }
    const RGB operator -(const RGB &p) { return RGB(r-p.r, g-p.g, b-p.b); }
    const RGB operator +(FLOAT_ARG p) { return RGB(r+p, g+p, b+p); }
    const RGB operator -(FLOAT_ARG p) { return RGB(r-p, g-p, b-p); }
    const RGB operator *(FLOAT_ARG p) { return RGB(r*p, g*p, b*p); }
    const RGB operator /(FLOAT_ARG p) { return RGB(r/p, g/p, b/p); }
    
    const unsigned char getR() const { return intensity(r); }
    const unsigned char getG() const { return intensity(g); }
    const unsigned char getB() const { return intensity(b); }
    
    const double power() const { return r+g+b; }
    
    void debug()
    {
//      printf("Returning %.2f %.2f %.2f\n", r,g,b);
    }
};

static class Fract
{
public:
     /* Nm vakiot kuuluvat Fract-luokalle, mutta kun ansi c++ ei salli
     * mritt static const doubleja luokalle. static const inttej vain.
     * GNU C++ selviisi kyll niist, mutta en jaksa katsella niit varoituksia...
     * Update: ovat nyt sittenkin tll. Koodista tulisi sotkuista ilman nit
     * luokan sisll. Poistan varoituksen kntjn optiolla.
     */
    static const double xmin = -2.0;
    static const double xmax = +2.0;
    static const double ymin = -1.3;
    static const double ymax = +1.3;
    static const double rmin = -2.7;
    static const double rmax = +0.5;
    static const double imin = -1.13;
    static const double imax = +1.13;
#if PROFILEMINMAX
    double xmi,xma, ymi,yma, rmi,rma, imi,ima;
#endif
public:
    Fract()
#if PROFILEMINMAX
    : xmi(xmax),xma(xmin),
              ymi(ymax),yma(ymin),
              rmi(rmax),rma(rmin),
              imi(imax),ima(imin)
#endif
    {
#if USEMANDMAP
        double ip = imin;
        const double iadd=(imax-imin) / YDIM;
        const double radd=(rmax-rmin) / XDIM;
        for(unsigned y=0; y<YDIM; ++y)
        {
            double rp = rmin;
            for(unsigned x=0; x<XDIM; ++x)
            {
                mandelmap[y][x] = ismandel(rp, ip);
                rp += radd;
            }
            ip += iadd;
        }
#endif
    }
    
private:
#if USEMANDMAP
    bool mandelmap[YDIM][XDIM];
#endif
    inline bool ismandel(FLOAT_ARG RP, FLOAT_ARG IP)
    {
        if(IP<imin || IP>imax
        || RP<rmin || RP>rmax)return false;
        register double r = RP, i = IP;
        for(register int c=0; ; )
        {
            register const double r2 = r*r;
            register const double i2 = i*i;
            
            if(r2+i2 >= 4)
            {
                /* Would never return back to <4. Not part of set. Exit. */
                return false;
            }
            
            i = r*i*2 + IP;
            r = r2-i2 + RP;
            
            if(++c >= MAXITER)
            {
                /* Tried enough. We are inside. */
                return true;
            }
        }
    }

    inline bool isjulia(FLOAT_ARG XP, FLOAT_ARG YP, FLOAT_ARG RP, FLOAT_ARG IP)
    {
#if SPHERETEST
        if((XP*XP + (RP-1.3)*(RP-1.3) + IP*IP) < 1.4)return true;
        if((XP*XP + (RP+1.3)*(RP+1.3) + IP*IP) < 1.4)return true;
        if(((XP-2.3)*(XP-2.3) + (RP)*(RP) + (IP-2.5)*(IP-2.5)) < 0.8)return true;
        if(((XP+1.6)*(XP+1.6) + (RP)*(RP) + (IP+1.2)*(IP+1.2)) < 0.6)return true;
        return false;
#else        
        if(wayout(XP,YP, RP,IP))return false;
#if USEMANDMAP
        if(!mandelmap[unsigned((IP-imin)*YDIM/(imax-imin))]
                     [unsigned((RP-rmin)*XDIM/(rmax-rmin))])return false;
#endif
        //if(((XP)*(XP) + (RP+1)*(RP+1) + (IP-0.6)*(IP-0.6)) < 1)return true;
        
        register double r = XP, i = YP;
        for(register int c=0; ; )
        {
            register const double r2 = r*r;
            register const double i2 = i*i;
            
            if(r2+i2 >= 4)
            {
                /* Would never return back to <4. Not part of set. Exit. */
                return false;
            }
            
            i = r*i*2 + IP;
            r = r2-i2 + RP;
            
            if(++c >= MAXITER)
            {
                /* Tried enough. We are inside. */
#if PROFILEMINMAX
                if(XP<xmi){xmi=XP;printf("New X min: %.6f\n", xmi);}
                if(YP<ymi){ymi=YP;printf("New Y min: %.6f\n", ymi);}
                if(RP<rmi){rmi=RP;printf("New R min: %.6f\n", rmi);}
                if(IP<imi){imi=IP;printf("New I min: %.6f\n", imi);}
                if(XP>xma){xma=XP;printf("New X max: %.6f\n", xma);}
                if(YP>yma){yma=YP;printf("New Y max: %.6f\n", yma);}
                if(RP>rma){rma=RP;printf("New R max: %.6f\n", rma);}
                if(IP>ima){ima=IP;printf("New I max: %.6f\n", ima);}
#endif
                return true;
            }
        }
#endif
    }
    
public:
    /* Paluuarvot: true=sisll, false=ulkona */
    inline bool hit(FLOAT_ARG x, FLOAT_ARG y, FLOAT_ARG r, FLOAT_ARG i)
    {
        return isjulia(x,y,r,i);
    }
    
    /* Paluuarvot: true=tosi ulkona, false=ehk sisll */
    bool wayout(FLOAT_ARG x, FLOAT_ARG y, FLOAT_ARG r, FLOAT_ARG i)
    {
        /* Mandelbrot-rajat */
        if(i<imin || i>imax
        || r<rmin || r>rmax)return true;

        /* Julia-rajat */
        if(x<xmin || x>xmax
        || y<ymin || y>ymax)return true;
        
        return false;
    }
} Fract;

static const double randang()
{
    return drand48() * 2 * M_PI;
}

class Light
{
public:
    XYZ pos, oldpos;
    RGB type;
    double rad;
    
    Light(const XYZ &p, const RGB &c, FLOAT_ARG r) : pos(p), type(c), rad(r)
    {
        oldpos = pos;
    }
};

static vector<Light> Lights;

static const double w = INITIAL_W; /* the time axis position */
#if SKYKLUDGE
static XYZ endpoint;
#endif

/* return value: false=mahiksia, true=ulos menossa */
static bool NoChances(const XYZ &beam, const XYZ &dir)
{
    if((beam.z < Fract.xmin && (beam.z+dir.z*MAXTRACE < Fract.xmin))
    || (beam.x < Fract.rmin && (beam.x+dir.x*MAXTRACE < Fract.rmin))
    || (beam.y < Fract.imin && (beam.y+dir.y*MAXTRACE < Fract.imin))
    || (w      < Fract.ymin              )
    
    || (beam.z > Fract.xmax && (beam.z+dir.z*MAXTRACE > Fract.xmax))
    || (beam.x > Fract.rmax && (beam.x+dir.x*MAXTRACE > Fract.rmax))
    || (beam.y > Fract.imax && (beam.y+dir.y*MAXTRACE > Fract.imax))
    || (w      > Fract.ymax              )
    )
        return true;

    return false;
}

static void SkipEmpty(XYZ &beam, const XYZ &dir, unsigned &step)
{
#if 0
    printf("z %.6f += %.6f\n"
           "x %.6f += %.6f\n"
           "y %.6f += %.6f\n",
           beam.z, dir.z,
           beam.x, dir.x,
           beam.y, dir.y);
#endif           
    #if 1
    
    double zspace = 0;
    double wspace = 0;
    double xspace = 0;
    double yspace = 0;

    if(beam.z < Fract.xmin && dir.z > 0)zspace = Fract.xmin - beam.z; else if(beam.z > Fract.xmax && dir.z < 0)zspace = beam.z - Fract.xmax;
    if(beam.x < Fract.rmin && dir.x > 0)xspace = Fract.rmin - beam.x; else if(beam.x > Fract.rmax && dir.x < 0)xspace = beam.x - Fract.rmax;
    if(beam.y < Fract.imin && dir.y > 0)yspace = Fract.imin - beam.y; else if(beam.y > Fract.imax && dir.y < 0)yspace = beam.y - Fract.imax;
    if(w      < Fract.ymin             )wspace = Fract.ymin - w     ; else if(w      > Fract.ymax             )wspace = w      - Fract.ymax;
    
    double minimi = sqrt(zspace*zspace + wspace*wspace + xspace*xspace + yspace*yspace) - BEAMSPEED;
    //double minimi = max(max(zspace,wspace), max(xspace,yspace)) - BEAMSPEED;
    
    #if 0
    for(unsigned l=0; l<Lights.size(); ++l)
    {
        double dist = beam.dist(Lights[l].pos);
        if(dist < minimi)minimi = dist;
    }
    #endif
    if(minimi >= BEAMSPEED)
    {
        double add = minimi / dir.length();
        step += int(add);
        beam += dir * add;
    }
    #endif
}

// Tarkoitus on muuttaa beamia, mutta ei dirri.
static void FocusHit(XYZ &beam, XYZ dir)
{
    /// Find the most exact position of hit.
    double prec = BEAMSPEED;
    if(prec > PRECISION)
    {
        /* Beam points to stuff. */
        beam -= dir;
        for(;;)
        {
            /* Beam points to space. */
            dir /= 2;
            prec /= 2;
            beam += dir;
            
            /* Check where does the beam point now. */            
            if(Fract.hit(beam.z, w, beam.x, beam.y))
            {
                /* Enough precision? */
                if(prec <= PRECISION)return;
                /* Back, want more precision */
                beam -= dir;
            }
            /* Continue */
        }
    }
    /* Done */
}

static void RelativeTrace(XYZ &beam, const XYZ &dir)
{
    /* Jos ollaan sisll, */
    if(Fract.hit(beam.z, w, beam.x, beam.y))
    {
        /* Kaivaudutaan ulos */
        do {
            beam -= dir;
        } while(Fract.hit(beam.z, w, beam.x, beam.y));
        /* Takaisin sislle */
        beam += dir;
    }
    else
    {
        /* Jos taas ulkona, trmtn sislle */
        for(unsigned max=0; ; )
        {
            beam += dir;
            if(Fract.hit(beam.z, w, beam.x, beam.y))break;
            
            if(++max >= MAXTRACE)
                return;
        }
    }
    /* Sen jlkeen etsitn tarkka osumakohta */
    FocusHit(beam, dir);
}

/*
    tracetypes:
    -1  = find light
     0  = find surface, tracetype = bouncecount
*/
static RGB RayTrace(const XYZ &startpoint, const XYZ &dir, int tracetype)
{
    XYZ beam = startpoint;
    
#if SKYKLUDGE
    endpoint.nowhere = false;
#endif
    unsigned step=0;    
    
    for(; ; ++step)
    {
        if(tracetype == -1)
        {
            /// Trying to find light?
            /// Then check if we are within the light source radius.
            for(unsigned l=0; l<Lights.size(); ++l)
            {
                double dist = beam.dist(Lights[l].pos);
                if(dist < Lights[l].rad)
                {
                    /// Inside the light source.
                	goto NoChances;
                }
            }
        }
        
    #if !SPHERETEST
        if(tracetype >= 0 && NoChances(beam, dir))
        {
        	goto NoChances;
        }
        if(!(step % (unsigned)(MAXTRACE/30)))
            SkipEmpty(beam, dir, step);
    #endif
    
        if(Fract.hit(beam.z, w, beam.x, beam.y))
        {
            if(tracetype < 0)
            {
                /// We were trying to find light.
                /// Instead, we hit a surface.
                /// So we found no light.
                return RGB AMBIENT;
            }
            
            FocusHit(beam, dir);
            
            XYZ randvec(0,0,1);
            if(dir.x==0.0)randvec.x=1;
            XYZ xdir = randvec.CrossProduct(dir);
            XYZ dir1 = dir.RotatedAround(xdir, DISTORTION);
            XYZ dir2 = dir1.RotatedAround(dir, DEG2RAD(240));
            XYZ dir3 = dir1.RotatedAround(dir, DEG2RAD(120));
            
            double surfacedistance = beam.dist(startpoint);
            
            XYZ end1 = startpoint + dir1 * (surfacedistance); RelativeTrace(end1, dir1);
            XYZ end2 = startpoint + dir2 * (surfacedistance); RelativeTrace(end2, dir2);
            XYZ end3 = startpoint + dir3 * (surfacedistance); RelativeTrace(end3, dir3);
            
            XYZ normal = (end2-end1).CrossProduct(end3-end1);
            normal.Normalize();

            /// We have now hit a solid surface.
            
            /// Declare the surface colour.
            static const RGB pigment PIGMENT;
            
            /// Declare the colour as it appears in the current lighting.            
            RGB result AMBIENT;
            
            /// Now check all the light sources
            for(unsigned l=0; l<Lights.size(); ++l)
            {
                XYZ live = Lights[l].pos - beam;
                live.Normalize();
                
                double effect = normal.DotProduct(live);
                if(effect <= 0)continue;
                if(DIRECTLIGHTMULTIPLIER != 0.0)
                {
                    /* Seek for something blocking the light, or then add it. */
                    XYZ tmp = live * BEAMSPEED;
                    result += RayTrace(beam+tmp, tmp, -1) * (DIRECTLIGHTMULTIPLIER * effect);
                }
                if(INDIRECTLIGHTMULTIPLIER != 0.0)
                {
                    /* Draw the light, whether it's seen or not */
                    /* This ignores all possible shadows. */
                    result += Lights[l].type * (effect * INDIRECTLIGHTMULTIPLIER);
                }
            }
            
            if(MIRRORLIGHTMULTIPLIER != 0.0)
            {
                XYZ reversevec = -dir;
                XYZ mirrored = reversevec.MirroredAround(normal);
                
                /*dir.print();
                printf("\nmirrored around "); normal.print();
                printf("\nis: "); mirrored.print();
                printf("\n\n");*/
                
                /* Find light in this mirror direction */
                result += RayTrace(beam+mirrored, mirrored, -1) * (MIRRORLIGHTMULTIPLIER);
                
                /* Find mirror surfaces too, if wanted */
                if(tracetype > 0)
                {
                    result += RayTrace(beam+mirrored, mirrored, tracetype - 1) * MIRRORLIGHTMULTIPLIER;
                }
            }
            
            /// The pigment affects all light sources.            
            result *= pigment;
            
            result.debug();
            return result;
        }

        if(step >= MAXTRACE)
        {
        	// No chances
        	break;
        }
        
        beam += dir;
    }
    
NoChances:
#if SKYKLUDGE
    endpoint = beam;
    endpoint.nowhere = true;
#endif
    /* Found *nothing* in this direction */
    RGB result(0,0,0);
#if SKYKLUDGE
    if(tracetype == -1)
    {
#endif
        XYZ yksdir(dir);
        yksdir.Normalize();
        /* Calculate some halo */
        for(unsigned l=0; l<Lights.size(); ++l)
        {
            XYZ live = Lights[l].pos - startpoint;
            live.Normalize();
            
            double effect = yksdir.DotProduct(live);
            if(effect <= 0)continue;
            unsigned potens = (tracetype<0) ? TRACEPOW : SKYPOW;
            result += Lights[l].type * pow(effect, potens);
        }
#if SKYKLUDGE
    }
#endif
    return result;
}

int main(int argc, const char *const *argv)
{
printf("  Heijastuslhteiden laskennan rekursioraja: %d\n"
       "  Peilimisesti heijastuneen valon kerroin: %.2f\n"
       "  Epsuoran valaistuksen (valon suuntaan perustuvan) kerroin: %.2f\n"
       "  Suoran valaistuksen (varjoihin perustuvan) kerroin: %.2f\n"
       "  Pinnan pigmentti: %s\n"
       "  Ambient-valo: %s\n"
       "\n"
       "Fraktaalin parametrit:\n"
       "  Iteraatiomr: %d\n"
       "  Akselit: mandelbrot{z,w}, julia{x,y} jossa w=%.2f\n"
       "\n"
       "Ja muut laskentaparametrit:\n"
       "  Kuvasuhde: %u/%u\n"
       "  Harppausvli: %g\n"
       "  Pinnan tarkkuus: %g\n"
       "  Kameran etisyys origosta: %g\n"
       "  Projektiopinnan etisyys kamerasta: %g\n"
       "  Maksimiharppausmr: %s (= %u)\n"
       "\n",
       MIRRORCOUNT, 
       MIRRORLIGHTMULTIPLIER,
       INDIRECTLIGHTMULTIPLIER,
       DIRECTLIGHTMULTIPLIER,
       PIGMENTSTR,
       AMBIENTSTR,
       MAXITER,
       w,
       XDIM, YDIM,
       BEAMSPEED,
       PRECISION,
       EYEDIST,
       PROJDIST,
       MAXTRACESTR,
       (unsigned)(MAXTRACE));

//	Lights.push_back(Light(XYZ(+0.0, -3.6, +1.0),   RGB(0.30, 0.10, 1.10), 0.3));   // Violetti taustavalo
//	Lights.push_back(Light(XYZ(-3.0, +3.3, -2.4),   RGB(1.00, 0.40, 0.10), 0.2));   // Punainen nurkkavalo
//	Lights.push_back(Light(XYZ(-1.0, -1.0, -2.0),   RGB(0.30, 0.70, 0.20), 0.1)); // Vaaleanvihre valo...
//	Lights.push_back(Light(XYZ(+0.0, +0.0, -2.6),   RGB(1.00, 1.00, 1.00), 1.0)); // Valkoinen spottiheijastusvalo
//	Lights.push_back(Light(XYZ(+2.0, +4.0, +1.0),   RGB(0.10, 0.60, 1.00), 0.4));   // Sininen nurkkavalo
//	Lights.push_back(Light(XYZ(+4.0, +0.0, -0.0),   RGB(0.30, 0.10, 1.10), 2.3));   // Sininen sivuvalo
//	Lights.push_back(Light(XYZ(+0.0, +0.0, -7.0),   RGB(0.30, 0.10, 1.10), 2.3));   // Sininen KESKIvalo

	Lights.push_back(Light(XYZ(-2.0, -0.0, -4.0), RGB(1.00,0.60,0.15), 0.1)); 
	Lights.push_back(Light(XYZ(+2.0, -2.0, -3.3), RGB(0.25,1.00,0.25), 0.1)); 
	Lights.push_back(Light(XYZ(+2.0, +1.6, -3.3), RGB(0.45,0.25,1.00), 0.1)); 
    
    int Parilliset = 1;
    int Parittomat = 1;
    
    for(int a=1; a<argc; ++a)
    {
        const char *s = argv[a];
        if(!strcmp(s, "2"))Parittomat = 0;
        if(!strcmp(s, "1"))Parilliset = 0;
    }

    if(Parilliset && !Parittomat) ALKUY = YDIM;
    else if(Parittomat && !Parilliset) ALKUY = 0;
    else ALKUY = 0;//(1024-YDIM)/4;

    for(double ax=1.4; ax<M_PI*4; ax+=M_PI / 2000)
    //double ax;
    //for(ax=0, w=-1.3; w<=1.3; w+=0.01)
    {
        //Matrix eyematrix(ax-M_PI, -sin(ax/2), -0.3 + sin(ax));
        //Matrix eyematrix(0.1, -0.7, 0);
        Matrix eyematrix(M_PI/2+ax, M_PI/4+ax, 0);
        //Matrix lightmatrix(ax*2, cos(ax)-0.9, cos(ax));
        Matrix lightmatrix = eyematrix;
        
        static unsigned imgind=2000;
        static const unsigned char tgaBuf[] = {0,0,2, 0,0, 0,0, 0,  0,0, 0,0, XDIM&255, XDIM>>8, YDIM&255, YDIM>>8, 24,32};
        char fnBuf[64];
        sprintf(fnBuf, "tga/mand%04u.tga", ++imgind);
        
        //if(imgind < 1603)continue;
        
        if((imgind & 1)==1 && !Parittomat)continue;
        if((imgind & 1)==0 && !Parilliset)continue;
        
        FILE *fp = fopen(fnBuf, "wb");
        fwrite(tgaBuf, 1, sizeof tgaBuf, fp);

        XYZ beam(0, 0, -EYEDIST);
        beam.Rotate(eyematrix);
	    beam.x -= 0.3;
        
        for(unsigned l=0; l<Lights.size(); ++l)
        {
            Lights[l].pos = Lights[l].oldpos;
            Lights[l].pos.Rotate(lightmatrix);
        }

        for(unsigned y=0; y<YDIM; ++y)
        {
            const double ydisp = (y / (double)(YDIM-1) - 0.5) * (PROJSIZEY*2.0);
            
            for(unsigned x=0; x<XDIM; ++x)
            {
                const double xdisp = (x / (double)(XDIM-1) - 0.5) * (PROJSIZEX*2.0);
                
                XYZ dir(xdisp,ydisp,PROJDIST);
                dir.SetLength(BEAMSPEED);
                dir.Rotate(eyematrix);
                
#if DISPLAYRENDER
                PSet(x,y, 15);
#endif
                RGB pix = RayTrace(beam, dir, MIRRORCOUNT);
#if SKYKLUDGE
                if(endpoint.nowhere)
                {
                    /* Point sky? Then draw a colourful sky */
                    
                    endpoint = beam + dir * MAXTRACE;
                    
                    for(unsigned l=0; l<Lights.size(); ++l)
                    {
                        double dist = pow(endpoint.dist(Lights[l].pos), 3.1);
                        if(dist<0.01)dist=0.01;
                        double power = 1.0 / dist;
                        pix += Lights[l].type * power;
                    }
                }
#endif
                const unsigned char R = pix.getR();
                const unsigned char G = pix.getG();
                const unsigned char B = pix.getB();
                
                putc(B, fp);
                putc(G, fp);
                putc(R, fp);
#if DISPLAYRENDER
                unsigned char c = (B*5/256)
                                + (R*5/256) * 5
                                + (G*9/256) * (5*5);
                PSet(x,y, 16 + c);
#endif
            }
        }
        
        fclose(fp);
        printf("Next\n"); fflush(stdout);
    }
    return 0;
}
